@@ -34,8 +34,12 @@ int xgetrandom(void *buf, unsigned buflen, unsigned flags)
 {
   int fd;
 
-#if CFG_TOYBOX_GETRANDOM
-  if (buflen == getrandom(buf, buflen, flags&~WARN_ONLY)) return 1;
+  // Linux keeps getrandom() in <sys/random.h> and getentropy() in <unistd.h>
+  // BSD/macOS only has getentropy(), but it's in <sys/random.h> (to be fair,
+  // they were there first). getrandom() and getentropy() both went into glibc
+  // in the same release (2.25 in 2017), so this test still works.
+#if __has_include(<sys/random.h>)
+  if (!getentropy(buf, buflen)) return 1;
   if (errno!=ENOSYS && !(flags&WARN_ONLY)) perror_exit("getrandom");
 #endif
   fd = xopen(flags ? "/dev/random" : "/dev/urandom",O_RDONLY|(flags&WARN_ONLY));
@@ -188,7 +192,7 @@ struct mtab_list *xgetmountlist(char *path)
 
 #endif
 
-#if defined(__APPLE__) || defined(__OpenBSD__)
+#if defined(__APPLE__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 
 #include <sys/event.h>
 
@@ -332,7 +336,7 @@ ssize_t xattr_fset(int fd, const char* name,
   return fsetxattr(fd, name, value, size, 0, flags);
 }
 
-#elif !defined(__OpenBSD__)
+#elif !defined(__FreeBSD__) && !defined(__OpenBSD__)
 
 ssize_t xattr_get(const char *path, const char *name, void *value, size_t size)
 {
@@ -457,7 +461,6 @@ static const struct signame signames[] = {
   // Non-POSIX signals that don't cause termination
   SIGNIFY(WINCH),
 };
-int signames_len = ARRAY_LEN(signames);
 
 #undef SIGNIFY
 
@@ -465,7 +468,6 @@ void xsignal_all_killers(void *handler)
 {
   int i;
 
-  if (!handler) handler = SIG_DFL;
   for (i = 1; signames[i].num != SIGCHLD; i++)
     if (signames[i].num != SIGKILL) xsignal(signames[i].num, handler);
 }
@@ -477,8 +479,8 @@ int sig_to_num(char *sigstr)
   char *s;
 
   // Numeric?
-  i = estrtol(sigstr, &s, 10);
-  if (!errno && !*s) return i;
+  offset = estrtol(sigstr, &s, 10);
+  if (!errno && !*s) return offset;
 
   // Skip leading "SIG".
   strcasestart(&sigstr, "sig");
@@ -512,7 +514,7 @@ char *num_to_sig(int sig)
   int i;
 
   // A named signal?
-  for (i=0; i<signames_len; i++)
+  for (i=0; i<ARRAY_LEN(signames); i++)
     if (signames[i].num == sig) return signames[i].name;
 
   // A real-time signal?
@@ -535,7 +537,7 @@ int dev_minor(int dev)
   return ((dev&0xfff00000)>>12)|(dev&0xff);
 #elif defined(__APPLE__)
   return dev&0xffffff;
-#elif defined(__OpenBSD__)
+#elif defined(__FreeBSD__) || defined(__OpenBSD__)
   return minor(dev);
 #else
 #error
@@ -548,7 +550,7 @@ int dev_major(int dev)
   return (dev&0xfff00)>>8;
 #elif defined(__APPLE__)
   return (dev>>24)&0xff;
-#elif defined(__OpenBSD__)
+#elif defined(__FreeBSD__) || defined(__OpenBSD__)
   return major(dev);
 #else
 #error
@@ -561,7 +563,7 @@ int dev_makedev(int major, int minor)
   return (minor&0xff)|((major&0xfff)<<8)|((minor&0xfff00)<<12);
 #elif defined(__APPLE__)
   return (minor&0xffffff)|((major&0xff)<<24);
-#elif defined(__OpenBSD__)
+#elif defined(__FreeBSD__) || defined(__OpenBSD__)
   return makedev(major, minor);
 #else
 #error
@@ -570,7 +572,7 @@ int dev_makedev(int major, int minor)
 
 char *fs_type_name(struct statfs *statfs)
 {
-#if defined(__APPLE__) || defined(__OpenBSD__)
+#if defined(__APPLE__) || defined(__FreeBSD__) || defined(__OpenBSD__)
   // macOS has an `f_type` field, but assigns values dynamically as filesystems
   // are registered. They do give you the name directly though, so use that.
   return statfs->f_fstypename;
@@ -625,26 +627,39 @@ int get_block_device_size(int fd, unsigned long long* size)
 }
 #endif
 
-// TODO copy_file_range
 // Return bytes copied from in to out. If bytes <0 copy all of in to out.
-// If consuemd isn't null, amount read saved there (return is written or error)
+// If consumed isn't null, amount read saved there (return is written or error)
 long long sendfile_len(int in, int out, long long bytes, long long *consumed)
 {
   long long total = 0, len, ww;
+  int try_cfr = 1;
 
   if (consumed) *consumed = 0;
-  if (in<0) return 0;
-  while (bytes != total) {
+  if (in>=0) while (bytes != total) {
     ww = 0;
     len = bytes-total;
-    if (bytes<0 || len>sizeof(libbuf)) len = sizeof(libbuf);
 
     errno = 0;
-#if CFG_TOYBOX_COPYFILERANGE
-    len = copy_file_range(in, 0, out, 0, bytes, 0);
+    if (try_cfr) {
+      if (bytes<0 || bytes>(1<<30)) len = (1<<30);
+      // glibc added this constant in git at the end of 2017, shipped 2018-02.
+      // Android's had the constant for years, but you'll get SIGSYS if you use
+      // this system call before Android U (2023's release).
+#if defined(__NR_copy_file_range) && !defined(__ANDROID__)
+      len = syscall(__NR_copy_file_range, in, 0, out, 0, len, 0);
 #else
-    ww = len = read(in, libbuf, len);
+      errno = EINVAL;
+      len = -1;
 #endif
+      if (len < 0 && errno == EINVAL) {
+        try_cfr = 0;
+
+        continue;
+      }
+    } else {
+      if (bytes<0 || len>sizeof(libbuf)) len = sizeof(libbuf);
+      ww = len = read(in, libbuf, len);
+    }
     if (len<1 && errno==EAGAIN) continue;
     if (len<1) break;
     if (consumed) *consumed += len;
@@ -681,7 +696,7 @@ int timer_settime(timer_t t, int flags, struct itimerspec *new, void *old)
 // glibc requires -lrt for linux syscalls, which pulls in libgcc_eh.a for
 // static linking, and gcc 9.3 leaks pthread calls from that breaking the build
 // These are both just linux syscalls: wrap them ourselves
-#elif !CFG_TOYBOX_HASTIMERS
+#elif defined(__GLIBC__)
 int timer_create_wrap(clockid_t c, struct sigevent *se, timer_t *t)
 {
   // convert overengineered structure to what kernel actually uses
